####
import click
from base import *
import hashlib
import sys
import requests
import os
import time

def md5(file):
    if not fs.exists(file):
        return "nohash"
    mm = hashlib.new("md5")
    data = fs.readfile(file,"b")
    mm.update(data)
    return mm.hexdigest()


platform = None

nbits = "64" if sys.maxsize > 2**32 else "32"
if sys.platform.startswith("win"):
    platform="windows"+nbits
elif sys.platform.startswith("linux"):
    platform="linux"+nbits
elif sys.platform.startswith("darwin"):
    platform="mac"


testmode = int(os.environ.get("ZERYNTH_TESTMODE",0))
if testmode == 2:
    # special testing mode
    zdir = "zerynth2_test" if platform.startswith("win") else ".zerynth2_test"
else:
    zdir = "zerynth2" if platform.startswith("win") else ".zerynth2"


class Env():
    pass

env = Env()

env.home      = fs.path(fs.homedir(),zdir)
env.cfg       = fs.path(env.home,"cfg")
env.tmp       = fs.path(env.home,"tmp")
env.sys       = fs.path(env.home,"sys")
env.vms       = fs.path(env.home,"vms")

if testmode==1:
    # local
    env.git_url   = os.environ.get("ZERYNTH_GIT_URL","http://localhost/git")
    env.backend   = os.environ.get("ZERYNTH_BACKEND_URL","http://localhost/v1")
    env.connector = os.environ.get("ZERYNTH_ADM_URL","http://localhost/v1")
    env.patchurl  = os.environ.get("ZERYNTH_PATCH_URL","http://localhost/installer")
    env.packurl   = os.environ.get("ZERYNTH_PACK_URL","http://localhost")
elif testmode==2:
    # CI
    env.git_url   = os.environ.get("ZERYNTH_GIT_URL","https://test.zerynth.com/git")
    env.backend   = os.environ.get("ZERYNTH_BACKEND_URL","https://test.zerynth.com/v1")
    env.connector = os.environ.get("ZERYNTH_ADM_URL","https://testapi.zerynth.com:444/v1" )
    env.patchurl  = os.environ.get("ZERYNTH_PATCH_URL","https://test.zerynth.com/installer")
    env.packurl   = os.environ.get("ZERYNTH_PACK_URL","https://test.zerynth.com")
else:
    # remote
    env.git_url ="https://backend.zerynth.com/git"
    env.backend="https://backend.zerynth.com/v1"
    env.connector="https://api.zerynth.com/v1"
    env.patchurl= os.environ.get("ZERYNTH_PATCH_URL","https://backend.zerynth.com/installer")
    env.packurl= os.environ.get("ZERYNTH_PACK_URL","https://backend.zerynth.com")

# dist directories
# env.dist      = fs.path(env.home,"dist",version)
# env.ztc       = fs.path(env.home,"dist",version,"ztc")
# env.libs      = fs.path(env.home,"dist",version,"libs")
# env.nest      = fs.path(env.home,"dist",version,"nest")
# env.stdlib    = fs.path(env.home,"dist",version,"stdlib")
# env.vhal      = fs.path(env.home,"dist",version,"vhal") 
# env.studio    = fs.path(env.home,"dist",version,"studio")
# env.docs      = fs.path(env.home,"dist",version,"docs")
# env.examples  = fs.path(env.home,"dist",version,"examples")
# env.devices   = fs.path(env.home,"dist",version,"devices")
# env.things    = fs.path(env.home,"dist",version,"things")
# env.idb       = fs.path(env.home,"dist",version,"idb")

env.dist_dir = lambda x: fs.path(env.home,"dist",x)
env.ztc_dir = lambda x: fs.path(x,"ztc")
env.lib_dir = lambda x: fs.path(x,"libs")
env.stdlib_dir = lambda x: fs.path(x,"stdlib")
env.studio_dir = lambda x: fs.path(x,"studio")
env.vhal_dir = lambda x: fs.path(x,"vhal")
env.devices_dir = lambda x: fs.path(x,"devices")
env.things_dir = lambda x: fs.path(x,"things")
env.docs_dir = lambda x: fs.path(x,"docs")
env.examples_dir = lambda x: fs.path(x,"examples")
env.idb_dir = lambda x: fs.path(x,"idb")


@cli.command(help="Install/update Zerynth.")
@click.argument("package")
@click.argument("dest")
def prepare_offline(package,dest):
    fs.rmtree(dest)
    fs.makedirs(dest)
    info("Unpacking",fs.filesize(package),"Mb...")    
    fs.untar(package,dest)
    info("Done")

@cli.command(help="Rename Folder.")
@click.argument("src")
@click.argument("dst")
def rename(src,dst):
    try:
        fs.move(src,dst)
    except Exception as e:
        fatal("Can't rename:",e)

@cli.command(help="Install/update Zerynth.")
def clear():
    tempdir = fs.get_named_tempdir("zerynth")
    fs.rmtree(tempdir)

@cli.command(help="Install/update Zerynth.")
@click.argument("patch")
@click.argument("repo")
@click.option("--offline", default=False, help="path to the directory holding offline packages")
@click.option("--community", default=False, help="version of a previous installation to copy community libraries from")
@click.option("--mute", flag_value=True, default=False,help="suppress stdout")
def install(patch, repo, offline, community, mute):
    if mute: set_output_filter(False)

    #### load repository
    db = fs.get_json(repo)
    version = db["version"] 
    patches = db["patches"]
    env.version = version

    if patch not in patches:
        critical("No such patch! Available:",patches)

    # calc total size
    tsize = 0
    for pack in db["packs"]:
        if pack.get("sys") and pack.get("sys")!=platform:
            # skip!
            continue
        tsize+=pack["size"]

    ofn = lambda x,y: "~"+str(x//1024)+"~"+str(y//1024) 
    # download packages
    if not offline:
        tempdir = fs.get_named_tempdir("zerynth")
        for pack in db["packs"]:
            if pack.get("sys") and pack.get("sys")!=platform:
                # skip!
                info("Skipping",pack["fullname"],ofn(pack["size"],tsize))
                continue
            fullname = pack["fullname"]
            url = pack["url"]
            if patch in pack["patches"]:
                packpatch = patch
                packhash = pack["hashes"][pack["patches"].index(patch)]
            else:
                packpatch = pack["patches"][-1]
                packhash = pack["hashes"][-1]
            if packhash=="-":
                #skip: removed from patch
                info("Skipping",pack["fullname"],ofn(pack["size"],tsize))
                continue

            url=url+"/"+packpatch
            info("Downloading package",fullname,ofn(pack["size"],tsize))
            outfile = fs.path(tempdir,fullname+".tar.xz")
            pack["file"] = outfile
            if fs.exists(outfile) and md5(outfile)==packhash:
                #already downloaded!
                continue
            for attempt in range(5,16,5):
                try:
                    hash = download_file(url,outfile)
                    info(hash,packhash)
                    if hash!=packhash:
                        info("Package seems corrupted, retrying...")
                        time.sleep(attempt)
                        continue
                    break
                except Exception as e:
                    warning("Exception while downloading",e)
            else:
                critical("Can't finish download!")

    else:
        for pack in db["packs"]:
            if pack.get("sys") and pack.get("sys")!=platform:
                # skip!
                continue
            pack["file"] = fs.path(offline,pack["fullname"]+"-"+version+".tar.xz")

    # install packages
    try:
        for pack in db["packs"]:
            if not pack.get("file"):
                info("Skipping",pack["fullname"],ofn(pack["size"],tsize))
                continue
            fullname = pack["fullname"]
            distpath = env.dist_dir(version)
            info("Installing package",fullname,ofn(pack["size"],tsize))
            if fullname.startswith("lib."):
                install_lib(pack,version,distpath)
            elif fullname.startswith("core."):
                install_core(pack,version,distpath)
            elif fullname.startswith("sys."):
                install_sys(pack,version,distpath)
            elif fullname.startswith("vhal."):
                install_vhal(pack,version,distpath)
            elif fullname.startswith("board."):
                install_device(pack,version,distpath)
            else:
                info("Skipping",pack["fullname"],ofn(pack["size"],tsize))
    except Exception as e:
       fatal("Can't install packages!",e)

    # set configuration
    fs.makedirs(env.cfg)
    version_file = fs.path(env.cfg,"config.json")
    patch_file = fs.path(env.cfg,"patches.json")
    try:
        patches = fs.get_json(patch_file)
    except:
        patches = {}
    patches[version]=patch
    try:
        fs.set_json(patches,patch_file)
    except Exception as e:
        fatal("Can't set configuration!",e)
    try:
        fs.set_json({"version":version},version_file)
    except Exception as e:
        fatal("Can't set configuration!",e)

    #clear temp
    fs.rmtree(fs.get_named_tempdir("zerynth"))

    #check community
    if community:
        info("Copying community libraries from version",community)
        pdistpath = env.dist_dir(community)
        plibdir = fs.path(env.lib_dir(pdistpath),"community")
        cdistpath = env.dist_dir(version)
        clibdir = fs.path(env.lib_dir(cdistpath),"community")
        if fs.exists(plibdir):
            fs.rmtree(clibdir)
            fs.copytree(plibdir,clibdir)

def install_lib(package,version,distpath):
    lib_src = fs.get_tempdir()
    
    fs.untarxz(package["file"],lib_src)
    doc_src = fs.path(lib_src,"docs")

    libsdir = env.lib_dir(distpath)
    flds = package["fullname"].split(".")
    namespace = flds[1]
    name = flds[2]
    repo = "official"
    lib_dst = fs.path(libsdir,repo,namespace,name)

    fs.rmtree(doc_src)
    fs.rmtree(lib_dst)
    fs.copytree(lib_src,lib_dst)
    fs.del_tempdir(lib_src)

def install_sys(package,version,distpath):
    sys_src = fs.get_tempdir()
    fs.untarxz(package["file"],sys_src)
    pk = fs.get_json(fs.path(sys_src,"package.json"))
    tdir = pk["targetdir"] # no need for newpython anymore
    
    if package["fullname"].startswith("sys.zerynth.browser-"):
        # remove package.json, not needed as a tool and interfere with nw.js -_-
        fs.rm_file(fs.path(sys_src,"package.json"))
    
    dst = fs.path(env.sys,tdir)
    fs.rmtree(dst)
    fs.copytree(sys_src,dst)
    fs.del_tempdir(sys_src)

def install_vhal(package,version,distpath):
    vhal_src = fs.get_tempdir()
    fs.untarxz(package["file"],vhal_src)
    pk = fs.get_json(fs.path(vhal_src,"package.json"))
    tdir = pk["targetdir"]
    dst = fs.path(env.vhal_dir(distpath),tdir)
    fs.rmtree(dst)
    fs.copytree(vhal_src,dst)
    fs.del_tempdir(vhal_src)

def install_device(package,version,distpath):
    dev_dst = fs.path(env.devices_dir(distpath),package["fullname"].split(".")[2])
    fs.rmtree(dev_dst)
    fs.makedirs(dev_dst)
    fs.untarxz(package["file"],dev_dst)

def install_core(package,version,distpath):
    if package["fullname"]=="core.zerynth.toolchain":
        dst = env.ztc_dir(distpath)
    elif package["fullname"]=="core.zerynth.studio":
        dst = env.studio_dir(distpath)
    elif package["fullname"]=="core.zerynth.stdlib":
        dst = env.stdlib_dir(distpath)
    else:
        # just in case
        dst = fs.path(distpath,package["fullname"].split(".")[2])
    fs.rmtree(dst)
    fs.untarxz(package["file"],dst)
    if package["fullname"] == "core.zerynth.studio":
        #change package.json for nw.js
        packjson = {
                      "name": "Zerynth Studio",
                      "main": "index.html",
                      "window": {
                        "frame": True,
                        "min_width": 800,
                        "min_height": 600
                      },
                      "version":env.version,
                      "user-agent":"ide/%ver/"+platform
                    }
        # no icon for mac, it's already in the browser
        if not platform.startswith("mac"):
            packjson["window"].update({"icon":"img/Logo512.png"})
        fs.set_json(packjson,fs.path(dst,"package.json"))
        

def download_file(url,out):
    local_filename = out
    r = requests.get(url, stream=True)
    with open(local_filename, 'wb') as f:
        fs.copyfileobj(r.raw, f)
    return md5(local_filename)

if __name__=="__main__":
    cli()
